#include <signal.h>
#include <io.h>
#include "types.h"
#include "hardware.h"
#include "global.h"
#include "timerb.h"

#define OUTMOD_MASK ((OUTMOD0)|(OUTMOD1)|(OUTMOD2))

/*
** TIMING_MARGIN defines the delay in clock cycles between current contents
** of TBR and start of the pulse. 100 s seems to be reasonable value.
*/

#define TIMING_MARGIN (1000UL)

typedef enum {IT_PULSESTART, IT_PULSEEND, IT_PAUSE} INTERVALTYPE_T;

extern volatile uint16 iflags;

static uint32 u32cyclesremaining = 0UL, u32newcycles = 0UL, u32pulseinterval = 0UL;
static uint16 u16pulses = 0;
static INTERVALTYPE_T itIntervalType = IT_PAUSE;

void StartPulse(uint32 u32length, uint16 u16start);
void NextPeriod(void);

void init_timerb(void)
{
    TBCCR0 = 0;
    TBCCTL0 = CM_0 | CCIS_0 | CLLD_0 |OUTMOD_0;
    //
    TBCCR1 = 0;
    TBCCTL1 = CM_0 | CCIS_0 | SCS | CCIS_0 | CLLD_0 | OUTMOD_0 | CCIE;
    //
    TBCTL = TBCLGRP_0 | CNTL_0 | TBSSEL_2 | ID_0 | MC_CONT | TBCLR;
}

interrupt(TIMERB1_VECTOR) TimerBInt(void)
{
    uint32 u32tmp;    

    switch(TBIV)
    {
        case TBIV_CCR1:
            if (u32cyclesremaining > 65535UL)
            {
                NextPeriod();
            }
            else
            {
                switch (itIntervalType)
                {
                    case IT_PULSESTART:
                        TBCCTL1 = (TBCCTL1 & (~OUTMOD_MASK)) | OUTMOD_5;
                        TBCCR1 += (uint16)u32cyclesremaining;

                        // Schedule a delay
                        itIntervalType = IT_PAUSE;
                        u32cyclesremaining = u32pulseinterval;
                        u16pulses--;
                        break;

                    case IT_PULSEEND:
                        TBCCTL1 = (TBCCTL1 & (~OUTMOD_MASK)) | OUTMOD_5;
                        // Schedule a delay
                        itIntervalType = IT_PAUSE;
                        u32cyclesremaining = u32pulseinterval;

                        if (u32cyclesremaining > 65536UL)
                        {
                            NextPeriod();
                        }
                        else if (u32cyclesremaining > TIMING_MARGIN)
                        {
                            // Schedule a delay
                            u32tmp = u32cyclesremaining - TIMING_MARGIN;
                            TBCCR1 += (uint16)u32tmp;
                            u32cyclesremaining -= u32tmp;
                        }

                        u16pulses--;
                        break;

                    case IT_PAUSE:
                        if (u16pulses > 0)
                        {
                            if (u32cyclesremaining > TIMING_MARGIN)
                            {
                                // Schedule a delay
                                u32tmp = u32cyclesremaining - TIMING_MARGIN;
                                TBCCR1 += (uint16)u32tmp;
                                u32cyclesremaining -= u32tmp;
                            }
                            else
                            {
                                // Schedule a new pulse
                                StartPulse(u32newcycles, TBCCR1 + (uint16)u32cyclesremaining);
                            }
                        }
                        else
                        {
                            TBCCTL1 = (TBCCTL1 & (~OUTMOD_MASK)) | OUTMOD_0;
                        }
                        break;
                    default:
                        nop();
                }

            }
            break;

        default:
            nop();
    }
}

void NextPeriod(void)
{
    uint32 u32tmp;
    uint32 u32cycles = 65536UL;

    u32tmp = u32cyclesremaining - u32cycles;
    if (u32tmp < TIMING_MARGIN)
    {
        u32cycles -= (TIMING_MARGIN - u32tmp);
    }
    TBCCR1 += (uint16)u32cycles;
    u32cyclesremaining -= u32cycles;
}

void SetPulseWidth(uint16 u16PulseWidth)
{
    uint16 u16start = TBR + TIMING_MARGIN;

    u32newcycles = ((uint32)u16PulseWidth)*10UL;
    u32pulseinterval = 100000UL*10UL;
    u16pulses = 16;

    StartPulse(u32newcycles, u16start);
}

void StartPulse(uint32 u32length, uint16 u16start)
{
    if (u32length < (65536UL-TIMING_MARGIN))
    {
        itIntervalType = IT_PULSEEND;
        TBCCR0 = u16start;
        TBCCR1 = u16start+(uint16)u32length;
        TBCCTL1 = (TBCCTL1 & (~OUTMOD_MASK)) | OUTMOD_7;
        u32cyclesremaining = 0UL;
    }
    else
    {
        itIntervalType = IT_PULSESTART;
        TBCCR1 = u16start;
        TBCCTL1 = (TBCCTL1 & (~OUTMOD_MASK)) | OUTMOD_1;
        u32cyclesremaining = u32length;
    }
}
